﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.MetaOptimization {
  [StorableClass]
  public class ParameterConfiguration : DeepCloneable, IParameterConfiguration, IStorableContent {
    [Storable]
    public string Filename { get; set; }

    [Storable]
    protected bool optimize;
    public bool Optimize {
      get { return optimize; }
      set {
        if (optimize != value) {
          optimize = value;
          if (optimize) {
            PopulateValueConfigurations();
          } else {
            this.ValueConfigurations.Clear();
          }
          OnOptimizeChanged();
        }
      }
    }

    [Storable]
    protected string parameterName;
    public string ParameterName {
      get { return parameterName; }
      set {
        if (parameterName != value) {
          parameterName = value;
        }
      }
    }

    protected Type parameterDataType;
    public Type ParameterDataType {
      get { return this.parameterDataType; }
    }

    protected IEnumerable<IItem> validValues;
    public IEnumerable<IItem> ValidValues {
      get { return validValues; }
      protected set {
        if (this.validValues != value) {
          this.validValues = value;
        }
      }
    }

    protected Type valueDataType;
    public Type ValueDataType {
      get { return valueDataType; }
      protected set { this.valueDataType = value; }
    }

    protected ICheckedValueConfigurationCollection valueConfigurations;
    public ICheckedValueConfigurationCollection ValueConfigurations {
      get { return this.valueConfigurations; }
      protected set { this.valueConfigurations = value; }
    }

    // todo: chooses one of the parameter configurations as actual value
    protected IValueConfiguration actualValueConfiguration;
    public IValueConfiguration ActualValueConfiguration {
      get { return this.actualValueConfiguration; }
      set { this.actualValueConfiguration = value; }
    }

    public ConstrainedValue ConstrainedValue {
      get { return actualValueConfiguration.ConstrainedValue; }
      set { actualValueConfiguration.ConstrainedValue = value; }
    }
    
    // store parameter reference only for name/description/image/...
    private IValueParameter parameter;

    #region Constructors and Cloning
    public ParameterConfiguration(string parameterName, IValueParameter valueParameter) {
      this.ParameterName = parameterName;
      this.parameterDataType = valueParameter.GetType();
      this.parameter = valueParameter;
      this.valueDataType = valueParameter.DataType;
      this.validValues = GetValidValues();
      this.ValueConfigurations = new CheckedValueConfigurationCollection(this.valueDataType);
      this.ActualValueConfiguration = new ValueConfiguration(valueParameter.Value, valueParameter.DataType);
      if (Optimize) {
        PopulateValueConfigurations();
      }
    }

    public ParameterConfiguration() { }
    [StorableConstructor]
    protected ParameterConfiguration(bool deserializing) { }
    protected ParameterConfiguration(ParameterConfiguration original, Cloner cloner)
      : base(original, cloner) {
      this.ParameterName = original.ParameterName;
      this.parameterDataType = original.parameterDataType;
      this.parameter = cloner.Clone(original.parameter);
      this.ValueDataType = original.ValueDataType;
      this.ValidValues = original.ValidValues.Select(x => cloner.Clone(x));
      this.ValueConfigurations = cloner.Clone(original.ValueConfigurations);
      this.ActualValueConfiguration = cloner.Clone(original.ActualValueConfiguration);
      this.Optimize = original.optimize;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterConfiguration(this, cloner);
    }
    #endregion

    private void PopulateValueConfigurations() {
      foreach (IItem validValue in this.validValues) {
        IItem val;
        if (actualValueConfiguration.ConstrainedValue.Value != null && actualValueConfiguration.ConstrainedValue.ValueDataType == validValue.GetType()) {
          val = actualValueConfiguration.ConstrainedValue.Value; // use existing value for that type (if available)
        } else {
          val = validValue;
        }
        this.ValueConfigurations.Add(new ValueConfiguration(val, val.GetType()), true);
      }
    }

    private IEnumerable<IItem> GetValidValues() {
      return ApplicationManager.Manager.GetInstances(valueDataType).Select(x => (IItem)x).OrderBy(x => x.ItemName).ToArray();
    }

    //void ValidValues_CheckedItemsChanged(object sender, Collections.CollectionItemsChangedEventArgs<Collections.IndexedItem<IItem>> e) {
    //  var oldItems = e.OldItems.Select(x => x.Value);
    //  var newItems = e.Items.Select(x => x.Value);
    //  var intersection = oldItems.Intersect(newItems);
    //  var removedItems = oldItems.Where(x => !intersection.Contains(x));
    //  var addedItems = newItems.Where(x => !intersection.Contains(x));

    //  foreach (var item in removedItems) {
    //    RemoveValueConfiguration(item);
    //  }
    //  foreach (var item in addedItems) {
    //    this.ValueConfigurations.Add(new ValueConfiguration(item, item.GetType()));
    //  }
    //}

    //private void RemoveValueConfiguration(IItem item) {
    //  IValueConfiguration vc = this.ValueConfigurations.Single(x => x.ValueDataType == item.GetType());
    //  this.ValueConfigurations.Remove(vc);
    //}

    #region INamedItem Properties
    public virtual string Name {
      get { return ParameterName; }
      set { throw new NotSupportedException(); }
    }
    public virtual string Description {
      get { return parameter.Description; }
      set { throw new NotSupportedException(); }
    }
    public virtual bool CanChangeDescription {
      get { return false; }
    }
    public virtual bool CanChangeName {
      get { return false; }
    }
    public virtual string ItemDescription {
      get { return parameter.Description; }
    }
    public virtual System.Drawing.Image ItemImage {
      get { return parameter.ItemImage; }
    }
    public virtual string ItemName {
      get { return parameter.ItemName; }
    }
    public virtual Version ItemVersion {
      get { return parameter.ItemVersion; }
    }
    #endregion

    #region Events
    public virtual event EventHandler NameChanged;
    protected virtual void OnNameChanged(object sender, EventArgs e) {
      var handler = NameChanged;
      if (handler != null) handler(sender, e);
    }

    public virtual event EventHandler<CancelEventArgs<string>> NameChanging;
    protected virtual void OnNameChanging(object sender, CancelEventArgs<string> e) {
      var handler = NameChanging;
      if (handler != null) handler(sender, e);
    }

    public virtual event EventHandler DescriptionChanged;
    protected virtual void OnDescriptionChanged(object sender, EventArgs e) {
      var handler = DescriptionChanged;
      if (handler != null) handler(sender, e);
    }

    public virtual event EventHandler ItemImageChanged;
    protected virtual void OnItemImageChanged(object sender, EventArgs e) {
      var handler = ItemImageChanged;
      if (handler != null) handler(sender, e);
    }

    public virtual event EventHandler ToStringChanged;
    protected virtual void OnStringChanged(object sender, EventArgs e) {
      var handler = ToStringChanged;
      if (handler != null) handler(this, e); // important to set 'this' as sender
    }

    public virtual event EventHandler OptimizeChanged;
    protected virtual void OnOptimizeChanged() {
      var handler = OptimizeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    #endregion

    public override string ToString() {
      return string.Format("{0}: {1}", ParameterName, ConstrainedValue.Value);
    }

    public static IParameterConfiguration Create(IParameterizedNamedItem parent, IParameter parameter) {
      if (parameter is IValueParameter) {
        IValueParameter valueParameter = parameter as IValueParameter;

        return new ParameterConfiguration(parameter.Name, valueParameter);
      }
      return null;
    }
  }
}
